/*** -*- Mode: Javascript; tab-width: 2;

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is jslib team code.
The Initial Developer of the Original Code is jslib team.

Portions created by jslib team are
Copyright (C) 2000 jslib team.  All
Rights Reserved.

Original Author: Pete Collins <pete@mozdev.org>
Contributor(s):

***/

// STATUS: WORK IN PROGRESS

var JS_LIB_UNINSTALL_LOAD_MSG = 
                  "JS_LIB_UNISTALL library not loaded:\n"
                + " \tTo load use: chrome://jslib/content/jslib.js\n"
                + " \tThen: include('chrome://jslib/content/io/uninstall.js');\n\n";

// Make sure jslib.js is loaded
if (typeof(JS_LIB_LOADED) == 'boolean')
{

// Ensure we are in a DOM window environment
if (typeof(window) == 'undefined') {
  JS_LIB_UNINSTALL_LOAD_MSG = "\n\nError: This library can only be used in a "
                              + "dom window environment\n\n";
  dump(JS_LIB_UNINSTALL_LOAD_MSG);
} else {
  // Load dependency lib files
  if (typeof(JS_FILE_LOADED) != 'boolean')
    include(jslib_file);

  if (typeof(JS_DIRUTILS_LOADED) != 'boolean')
    include(jslib_dirutils);

  if (typeof(JS_DIR_LOADED) != 'boolean')
    include(jslib_dir);

  if (typeof(JS_FILEUTILS_LOADED) != 'boolean')
    include(jslib_fileutils);

  if (typeof(JS_COMMONWINDOW_LOADED) != 'boolean')
    include(jslib_window);

  // GLOBALS
  const JS_LIB_UNINSTALL_LOADED   = true;
  const JS_LIB_UNINSTALL_FILE     = 'uninstall.js';

  // CONSTRUCTOR
  function Uninstall(aPackageName) {

    if (!aPackageName) {
      jslibError(null, "Please enter a package name to uninstall",
               "NS_ERROR_XPC_NOT_ENOUGH_ARGS", JS_LIB_UNINSTALL_FILE);
      throw - Components.results.NS_ERROR_XPC_NOT_ENOUGH_ARGS;
    } else {
      this.mName = aPackageName;
      this.init();
    }

    this.gRDF = Components.classes['@mozilla.org/rdf/rdf-service;1']
                  .getService(Components.interfaces.nsIRDFService);

    this.gDirService = Components.classes["@mozilla.org/file/directory_service;1"]
                         .getService(Components.interfaces.nsIProperties);
  }

  /*********** UNINSTALL ***************/
  Uninstall.prototype = {

    mName           : null,
    mLastURL        : null,
    mLastChromeURL  : null,
    CHRM_REG_CID    : "@mozilla.org/chrome/chrome-registry;1",
    SKIN            : null,
    JAR             : null,
    CHROME_NS       : "http://www.mozilla.org/rdf/chrome#",
    filesToDelete   : [],
    gRDF            : null,
    gDirService     : null,

  init  : function () 
  {
    this.SKIN   = this.mName+"/";
    this.JAR    = this.mName+".jar";
  },

  closeOpenWindows : function () 
  {
    jslibDebug(opener.location);
    jslibDebug(opener._content.location);
    this.mLastChomeURL = opener.location.toString();
    this.mLastURL = opener._content.location.toString();
    var win = new CommonWindow();
    var openWindows = win.openWindows;
    for (var i=0; i<openWindows.length; i++)  {
      if (/chrome:\/\/jslib\/content\/install\/uninstall.xul/.test(openWindows[i].location)) {
        jslibPrint("Skipping: "+openWindows[i].location);
        continue;
      }
      openWindows[i].close();
    }

    return;
  },

  launchLastWindow : function () 
  {
    window.openDialog(this.mLastChomeURL, "_blank", "chrome,all,dialog=no", this.mLastURL);
  },

  uninstall : function () 
  {
    jslibDebug(opener.location);
    jslibDebug(opener._content.location);
    this.mLastChomeURL = opener.location.toString();
    this.mLastURL = opener._content.location.toString();
    this.closeOpenWindows(); 
  },

  finish : function () 
  {
    this.launchLastWindow();
  },

  uninstallPackage : function () 
  {
    this.doUninstall();
    return this.finishedDialog();
  },

  removeRDFEntries : function () 
  {
    var rv = Components.results.NS_OK;
    rv = this.removeOverlayEntry();
    // rv = this.removeAllPkgEntry();
    return rv;
  },

  clearMenu : function () 
  {
    jslibDebug('clearMenu: '+this.mName+'\n');
    var menu = document.getElementById(this.mName);
    if (!menu) 
      return null;
    jslibDebug(menu);
    var rv;
    var sep  = document.getElementById(this.mName+"_separator");
    rv = sep.setAttribute("style", "display: none;");
    rv = menu.setAttribute("style", "display: none;");
    return rv;
  },

  finishedDialog : function () 
  {
    const finish_msg  = this.mName+" is now uninstalled . . .\n";
    this.clearMenu();
    return alert(finish_msg);
  },

iterateContainer : function(ds, resource, callback)
{
  try {
    var container = Components.classes["@mozilla.org/rdf/container;1"].
                        createInstance(Components.interfaces.nsIRDFContainer);
    container.Init(ds, resource);
  }
  catch (ex){ jslibDebug(ex); return; }

  var elements = container.GetElements();
  while (elements.hasMoreElements()){
    var element = elements.getNext();
    callback(resource, element, this);
  }
},

doUninstall : function()
{
  var ioService = Components.classes['@mozilla.org/network/io-service;1'].
                    getService(Components.interfaces.nsIIOService);

  this.filesToDelete = [];

  // scan through chrome.rdf and find all references to the package and remove them.
  var chromeDir = this.gDirService.get("AChrom", Components.interfaces.nsIFile);
  var chromeRdfFile = chromeDir.clone();
  chromeRdfFile.append("chrome.rdf");
  var chromeUrl = ioService.newFileURI(chromeRdfFile).spec;
  var chromeDS = this.gRDF.GetDataSourceBlocking(chromeUrl);

  var packagename = this.mName;

  // remove package from content
  var rootseq = this.gRDF.GetResource("urn:mozilla:package:root");
  var packres = this.gRDF.GetResource("urn:mozilla:package:" + packagename);
  this.doUninstallFromContents(chromeDS, chromeDir, rootseq, packres);

  // remove package from skin

  var provider = "skin";

  var handleSkinLocaleList = function(container, skinLocale, uninstallObj)
  {
    var rootseq = chromeDS.GetTarget(skinLocale,
                    uninstallObj.gRDF.GetResource(uninstallObj.CHROME_NS + "packages"),true);
    rootseq.QueryInterface(Components.interfaces.nsIRDFResource);

    var skinLocaleName = chromeDS.GetTarget(skinLocale,
          uninstallObj.gRDF.GetResource(uninstallObj.CHROME_NS + "name"),true);

    skinLocaleName.QueryInterface(Components.interfaces.nsIRDFLiteral);
    var skinLocaleRes = uninstallObj.gRDF.GetResource("urn:mozilla:" + provider + ":" +
                          skinLocaleName.Value + ":" + packagename);

    uninstallObj.doUninstallFromContents(chromeDS, chromeDir, rootseq, skinLocaleRes);
  };

  var packreslist = this.gRDF.GetResource("urn:mozilla:skin:root");
  this.iterateContainer(chromeDS, packreslist, handleSkinLocaleList);

  provider = "locale";

  packreslist = this.gRDF.GetResource("urn:mozilla:locale:root");
  this.iterateContainer(chromeDS, packreslist, handleSkinLocaleList);

  if (chromeDS instanceof Components.interfaces.nsIRDFRemoteDataSource)
    chromeDS.Flush();

  // all files are deleted at the end of the process
  for (var t=0; t<this.filesToDelete.length; t++){
    // ignore errors since it doesn't matter if a file could not be found, and
    // non-empty directories should not be deleted.
    try {
      var file = this.filesToDelete[t];
      if (file.exists()) file.remove(false);
    }
    catch (ex){ jslibDebug(ex); }
  }
},

/**
 * Perform an uninstallation given a contents.rdf datasource
 *   aChromeDS - chrome.rdf datasource
 *   chromeDir - the chrome directory
 *   rootseq   - root sequence
 *   packres   - packagename as a resource
 */
doUninstallFromContents : function(chromeDS, chromeDir, rootseq, packres)
{
  var baseUrlPred = this.gRDF.GetResource(this.CHROME_NS + "baseURL")
  var baseUrl = chromeDS.GetTarget(packres, baseUrlPred, true);
  if (baseUrl instanceof Components.interfaces.nsIRDFLiteral){
    try {
      var ds = this.gRDF.GetDataSourceBlocking(baseUrl.Value + "contents.rdf");
    }
    catch (ex){ jslibDebug(ex); return; }

    this.removeFilesFromUninstallInfo(ds, packres);
    this.removeAllOverlays(ds, chromeDir, "overlays");
    this.removeAllOverlays(ds, chromeDir, "stylesheets");
    this.removeFromChrome(chromeDS, rootseq, packres);
  }
},

/**
 * Remove the files that are listed in the uninstallInfo section
 * of the contents.rdf
 */
removeFilesFromUninstallInfo : function(aDS, node)
{
  var pred = this.gRDF.GetResource(this.CHROME_NS + "uninstallInfo");
  var uninstallInfo = aDS.GetTarget(node,pred,true);
  if (uninstallInfo){
    this.iterateContainer(aDS, uninstallInfo, this.removeFile);
  }
},

/**
 * Remove a file relative to the Mozilla directory.
 */
removeFile : function(container, filename, uninstallObj)
{
  if (!(filename instanceof Components.interfaces.nsIRDFLiteral)) return;
  filename = filename.Value;

  var filekey;
  var colonIdx = filename.indexOf(":");
  if (colonIdx >= 0){
    filekey = filename.substring(0,colonIdx);
    filename = filename.substring(colonIdx + 1);
  }
  else {
    filekey = "CurProcD";
  }

  var file;
  try {
     file = uninstallObj.gDirService.get(filekey, Components.interfaces.nsIFile);
  } catch (ex) { return; }

  var fileparts = filename.split("/");
  for (var t=0; t<fileparts.length; t++){
    file.append(fileparts[t]);
  }

  jslibDebug("Uninstall ---- Removing " + file.path + "\n");

  uninstallObj.filesToDelete.push(file);
},

/**
 * Remove all of the overlays referenced in a contents.rdf file.
 */
removeAllOverlays : function(aDS, chromeDir, overlayType)
{
  var iterateOverlays = function(container, overlayFile, uninstallObj)
  {
    if ((container instanceof Components.interfaces.nsIRDFResource) &&
        (overlayFile instanceof Components.interfaces.nsIRDFLiteral)){
      uninstallObj.removeOverlay(container, overlayFile, chromeDir, overlayType);
    }
  }

  var iterateOverlaids = function(container, overlaidFile, uninstallObj)
  {
    uninstallObj.iterateContainer(aDS, overlaidFile, iterateOverlays);
  }

  var oroot = this.gRDF.GetResource("urn:mozilla:" + overlayType);
  this.iterateContainer(aDS, oroot, iterateOverlaids);
},

/**
 * Remove an overlay from the overlayinfo.
 */
removeOverlay : function(overlaidFile, overlayFile, chromeDir, overlayType)
{
  var overlayItems = this.splitURL(overlaidFile.Value);

  var overlayRdfFile = chromeDir.clone();
  overlayRdfFile.append("overlayinfo");
  overlayRdfFile.append(overlayItems.packagename);
  overlayRdfFile.append(overlayItems.provider);
  overlayRdfFile.append(overlayType + ".rdf");

  var ioService = Components.classes['@mozilla.org/network/io-service;1'].
                    getService(Components.interfaces.nsIIOService);
  var overlayRdfUrl = ioService.newFileURI(overlayRdfFile).spec;
  var dsource = this.gRDF.GetDataSourceBlocking(overlayRdfUrl);

  try {
    var container = Components.classes["@mozilla.org/rdf/container;1"].
                      createInstance(Components.interfaces.nsIRDFContainer);
    container.Init(dsource, overlaidFile);
    jslibDebug("Uninstall ---- Uncontain Overlay " + this.RDFGetValue(overlayFile) +
         " from " + this.RDFGetValue(overlaidFile) + "\n");
    container.RemoveElement(overlayFile, true);
  }
  catch (ex) { jslibDebug(ex); }

  if (dsource instanceof Components.interfaces.nsIRDFRemoteDataSource)
    dsource.Flush();
},

/**
 * split a chrome URL into component parts.
 *
 * The algorithm was taken from mozilla/rdf/chrome/src/nsChromeRegistry.cpp
 */
splitURL : function(url)
{
  if (url.indexOf("chrome://")) return null;

  var packagename = url.substring(9);
  var slashidx = packagename.indexOf("/");
  if (slashidx == -1) return null;

  var provider = packagename.substring(slashidx + 1);
  var packagename = packagename.substring(0,slashidx);
 
  slashidx = provider.indexOf("/");
  if (slashidx >= 0){
    provider = provider.substring(0,slashidx);
  }

  return {
    packagename: packagename,
    provider: provider
  };
},

/**
 * Useful debugging function to convert an nsIRDFNode into a string.
 */
RDFGetValue : function(node)
{
  return ((node instanceof Components.interfaces.nsIRDFResource) ? node.Value :
          ((node instanceof Components.interfaces.nsIRDFLiteral) ? node.Value : ""));
},

/**
 * Remove references to a package from chrome.rdf.
 */
removeFromChrome : function (dsource, rootseq, packres) 
{
  var packresnode = packres.QueryInterface(Components.interfaces.nsIRDFNode);

  try {
    var container = Components.classes["@mozilla.org/rdf/container;1"].
                      createInstance(Components.interfaces.nsIRDFContainer);
    container.Init(dsource, rootseq);
    jslibDebug("Uninstall ---- Uncontain " + packres.Value + " from " +
               rootseq.Value + "\n");
    container.RemoveElement(packresnode, true);
  }
  catch (ex) { jslibDebug(ex); }

  var arcs = dsource.ArcLabelsOut(packres);

  while(arcs.hasMoreElements()) {
    var arc = arcs.getNext();
    
    var prop = arc.QueryInterface(Components.interfaces.nsIRDFResource);

    var targets = dsource.GetTargets(packres, prop, true);

    while (targets.hasMoreElements()) {
      var target = targets.getNext();

      var targetNode = target.QueryInterface(Components.interfaces.nsIRDFNode);
      jslibDebug("Uninstall ---- Unassert [" + packres.Value + " , " +
            prop.Value + " , " + this.RDFGetValue(target) + "]\n");
      dsource.Unassert(packres, prop, targetNode);
    }
  }
}

  /*********** UNINSTALL ***************/
  } // END CLASS
  jslibDebug('*** load: '+JS_LIB_UNINSTALL_FILE+' OK');

} // END BLOCK DOM WINDOW CHECK  

// END BLOCK JS_LIB_LOADED CHECK 
// If jslib base library is not loaded, dump this error.
} else {
  dump(JS_LIB_UNINSTALL_LOAD_MSG);
}
